/*
 * linux/arch/arm/mach-sa1100/jornada720.c
 */

#include <linux/init.h>
#include <linux/kernel.h>
#include <linux/tty.h>
#include <linux/delay.h>

#include <asm/hardware.h>
#include <asm/setup.h>

#include <asm/mach/arch.h>
#include <asm/mach/map.h>
#include <asm/mach/serial_sa1100.h>
#include <linux/serial_core.h>

#include "generic.h"
#include "sa1111.h"

#include <linux/wait.h>          /* for wait_queue_head_t, used in h3600_ts.h */
#include <linux/types.h>         /* for u_char, used in h3600_ts.h */
#include <linux/h3600_ts.h>      /* for TS_RET */


#define JORTUCR_VAL	0x20000400
#define JORSKCR_INIT	0x00002081	/* Turn off VCO to enable PLL, set Ready En and enable nOE assertion from DC */
#define JORSKCR_RCLK	0x00002083	/* Add turning on RCLK to above */
#define JORSKCR_VAL	0x0000001B	/* sets the 1101 control register to on */

#define MCU_GetBatteryData  0xc0
#define MCU_GetScanKeyCode	0x90
#define MCU_GetTouchSamples	0xa0
#define MCU_SetContrast     0xD1
#define MCU_SetBrightness   0xD3
#define MCU_TxDummy         0x11
#define MCU_ErrorCode       0x00
#define SA1100_PPDR_LFCLK 	0x400

#define mcu_read() mcu_byte(MCU_TxDummy)

static char mcu_invert[256];

static int mcu_byte(int arg_data)
{
int i;

	while ((Ser4SSSR & SSSR_TNF) == 0);
	i = 0;
	while ((GPLR & 0x400) && i++ < 400000)
		; /* wait for MCU */
	if (i >= 400000) {
		printk("mcu_byte: timed out\n");
        return -1;
    }
	Ser4SSDR = mcu_invert[arg_data] << 8;
	udelay(100);
	while ((Ser4SSSR & SSSR_RNE) == 0);
	i = Ser4SSDR;
	if (i > 0xff)
		printk("mcu_byte: read %x\n", i);
	return mcu_invert[ i & 0xff ];
}

static void mcu_start(int arg_data)
{
int i;

	jornada720_mcu_init();
	GPCR = 0x02000000; /* clear -> enable */
#if 0
	udelay(100);
#endif
	i = mcu_byte(arg_data);
	if (i != MCU_TxDummy)
		printk("mcu_start: [%x] bad start %x\n", arg_data, i);
}

static void mcu_end(void)
{
#if 0
	udelay(100);
#endif
	GPSR = 0x02000000; /* set */
}

void jornada720_mcu_init(void)
{
int i;
static int initialized = 0;

	if (initialized)
		return;
printk ("[%s:%d] %s ENTRY\n", __FILE__, __LINE__, __FUNCTION__);
	initialized = 1;
	PPSR &= ~(SA1100_PPDR_LFCLK | 0x80 | 0x10);
	udelay(500);
printk ("[%s:%d] %s PPSR reset\n", __FILE__, __LINE__, __FUNCTION__);
	PPDR |= SA1100_PPDR_LFCLK | 0x80 | 0x10;
	udelay(500);
printk ("[%s:%d] %s after reset\n", __FILE__, __LINE__, __FUNCTION__);
	/* Take the MCU out of reset mode */
	PPSR |= SA1100_PPDR_LFCLK;
	udelay(500);
	printk ("now wait after reset done\n");
	for (i = 0; i <= 255; i++)
		mcu_invert[i] = ( ((0x80 & i) >> 7) | ((0x40 & i) >> 5)
		  | ((0x20 & i) >> 3) | ((0x10 & i) >> 1) | ((0x08 & i) << 1)
		  | ((0x04 & i) << 3) | ((0x02 & i) << 5) | ((0x01 & i) << 7) );

	GPSR = 0x02000000; /* set */
	GPDR |= 0x02000000; /* GPIO(25) low -> enable MCU */
	Ser4SSCR0 = 0x0307;
	Ser4MCCR0 = 0;
	Ser4SSSR = 0;	/* remove any rcv overrun errors */
	Ser4SSCR1 = 0x18;
	Ser4SSCR0 = 0x0387;

	while (Ser4SSSR & SSSR_RNE)
		i = Ser4SSDR;		/* drain any data already there */
}

void jornada_contrast(int arg_contrast)
{
	mcu_start(MCU_SetContrast);
	mcu_byte(arg_contrast);
	mcu_end();
}

void jornada720_battery(void)
{
	mcu_start(MCU_GetBatteryData);
	printk("get2 bat %x\n", mcu_read());
	printk("get3 bat %x\n", mcu_read());
	printk("get4 bat %x\n", mcu_read());
	mcu_end();
}

int jornada720_getkey(unsigned char *data, int size)
{
int this_count, i, j;

	mcu_start(MCU_GetScanKeyCode);
	this_count = mcu_read();	/* Flush any pending input. */
	i = this_count;
	while (i-- > 0) {
		j = mcu_read();
		if (size-- > 0)
			*data++ = j;
	}
	mcu_end();
	return this_count;
}

void jornada720_gettouch(void *argret)
{
int x1, x2, x3, y1, y2, y3;
int i;
TS_RET *ret = (TS_RET *)argret;

	mcu_start(MCU_GetTouchSamples);
	x1 = mcu_read();
	x2 = mcu_read();
	x3 = mcu_read();
	i = mcu_read();
	x1 |= (i & 3) << 8;
	x2 |= (i & 0xc) << 6;
	x3 |= (i & 0x30) << 4;
	y1 = mcu_read();
	y2 = mcu_read();
	y3 = mcu_read();
	i = mcu_read();
	y1 |= (i & 3) << 8;
	y2 |= (i & 0xc) << 6;
	y3 |= (i & 0x30) << 4;
	mcu_end();
	ret->y = x2;
	ret->x = y2;
    ret->pressure = ( (GPLR & GPIO_JORNADA720_MOUSE) == 0);
}

static int __init jornada720_init(void)
{
	GPDR |= GPIO_GPIO20;
	TUCR = JORTUCR_VAL;	/* set the oscillator out to the SA-1101 */

	GPSR = GPIO_GPIO20;
	udelay(1);
	GPCR = GPIO_GPIO20;
	udelay(1);
	GPSR = GPIO_GPIO20;
	udelay(20);
	SKCR = JORSKCR_INIT;	/* Turn on the PLL, enable Ready and enable nOE assertion from DC */
	mdelay(100);

	SKCR = JORSKCR_RCLK;	/* turn on the RCLOCK */
	SMCR = 0x35;	/* initialize the SMC (debug SA-1111 reset */
	PCCR = 0;	/* initialize the S2MC (debug SA-1111 reset) */

	/* LDD4 is speaker, LDD3 is microphone */
	PPSR &= ~(PPC_LDD3 | PPC_LDD4);
	PPDR |= PPC_LDD3 | PPC_LDD4;

	sa1111_init();
	/* initialize extra IRQs */
	sa1111_init_irq(JORNADA720_SA1111_GPIO);	/* chained on GPIO 1 */
	jornada720_mcu_init();

	return 0;
}

__initcall(jornada720_init);

static void __init
fixup_jornada720(struct machine_desc *desc, struct param_struct *params,
		 char **cmdline, struct meminfo *mi)
{
	SET_BANK( 0, 0xc0000000, 32*1024*1024 );
	mi->nr_banks = 1;
}

static struct map_desc jornada720_io_desc[] __initdata = {
 /* virtual     physical    length      domain     r  w  c  b */
  { 0xe8000000, 0x00000000, 0x02000000, DOMAIN_IO, 1, 1, 0, 0 }, /* Flash bank 0 */
  { 0xf0000000, 0x48000000, 0x00100000, DOMAIN_IO, 1, 1, 0, 0 }, /* Epson registers */
  { 0xf1000000, 0x48200000, 0x00100000, DOMAIN_IO, 1, 1, 0, 0 }, /* Epson frame buffer */
  { 0xf4000000, 0x40000000, 0x00100000, DOMAIN_IO, 1, 1, 0, 0 }, /* SA-1111 */
  LAST_DESC
};

static void __init jornada720_map_io(void)
{
	sa1100_map_io();
	iotable_init(jornada720_io_desc);

	sa1100_register_uart(0, 3);
	sa1100_register_uart(1, 1);
}

MACHINE_START(JORNADA720, "HP Jornada 720")
	BOOT_MEM(0xc0000000, 0x80000000, 0xf8000000)
	BOOT_PARAMS(0xc0000100)
	FIXUP(fixup_jornada720)
	MAPIO(jornada720_map_io)
	INITIRQ(sa1100_init_irq)
MACHINE_END
